## {{ versiondata.version }} ({{ versiondata.date }})
{% for section, _ in sections.items() %}
{% set anchor = "#" * underlines[0] %}{% if section %}{{ anchor }} {{section}}
{% set anchor = "#" * underlines[1] %}

{% endif %}

{% if sections[section] %}
{% for category, val in definitions.items() if category in sections[section]%}
{{ anchor }} {{ definitions[category]['name'] }}

{% if definitions[category]['showcontent'] %}
{% for text, values in sections[section][category].items() %}
- {{ text }} ({{ values|join(', ') }})
{% endfor %}

{% else %}
- {{ sections[section][category]['']|join(', ') }}

{% endif %}
{% if sections[section][category]|length == 0 %}
No significant changes.

{% else %}
{% endif %}
{% endfor %}
{% else %}
No significant changes.

{% endif %}
{% endfor %}
---
{# 
This comment adds one more newline at the end of the rendered newsfile content. 
In this way the there are 2 newlines between the latest release and the previous release content.
#}
