# DO NOT MANUALLY EDIT THIS FILE!
# This file was automatically generated by `nox -s generate-stubs`

from hikari import api as api
from hikari import applications as applications
from hikari import events as events
from hikari import files as files
from hikari import impl as impl
from hikari import interactions as interactions
from hikari import snowflakes as snowflakes
from hikari import undefined as undefined
from hikari._about import __author__ as __author__
from hikari._about import __ci__ as __ci__
from hikari._about import __copyright__ as __copyright__
from hikari._about import __coverage__ as __coverage__
from hikari._about import __discord_invite__ as __discord_invite__
from hikari._about import __docs__ as __docs__
from hikari._about import __email__ as __email__
from hikari._about import __git_sha1__ as __git_sha1__
from hikari._about import __issue_tracker__ as __issue_tracker__
from hikari._about import __license__ as __license__
from hikari._about import __maintainer__ as __maintainer__
from hikari._about import __url__ as __url__
from hikari._about import __version__ as __version__
from hikari.applications import Application as Application
from hikari.applications import ApplicationContextType as ApplicationContextType
from hikari.applications import ApplicationFlags as ApplicationFlags
from hikari.applications import ApplicationIntegrationType as ApplicationIntegrationType
from hikari.applications import ApplicationRoleConnectionMetadataRecord as ApplicationRoleConnectionMetadataRecord
from hikari.applications import (
    ApplicationRoleConnectionMetadataRecordType as ApplicationRoleConnectionMetadataRecordType,
)
from hikari.applications import AuthorizationApplication as AuthorizationApplication
from hikari.applications import AuthorizationInformation as AuthorizationInformation
from hikari.applications import ConnectionVisibility as ConnectionVisibility
from hikari.applications import OAuth2AuthorizationToken as OAuth2AuthorizationToken
from hikari.applications import OAuth2ImplicitToken as OAuth2ImplicitToken
from hikari.applications import OAuth2Scope as OAuth2Scope
from hikari.applications import OwnApplicationRoleConnection as OwnApplicationRoleConnection
from hikari.applications import OwnConnection as OwnConnection
from hikari.applications import OwnGuild as OwnGuild
from hikari.applications import PartialOAuth2Token as PartialOAuth2Token
from hikari.applications import Team as Team
from hikari.applications import TeamMember as TeamMember
from hikari.applications import TeamMembershipState as TeamMembershipState
from hikari.applications import TokenType as TokenType
from hikari.audit_logs import *
from hikari.auto_mod import *
from hikari.channels import *
from hikari.colors import *
from hikari.colours import *
from hikari.commands import *
from hikari.components import *
from hikari.embeds import *
from hikari.emojis import *
from hikari.errors import *
from hikari.events.base_events import Event as Event
from hikari.events.base_events import ExceptionEvent as ExceptionEvent
from hikari.events.channel_events import *
from hikari.events.guild_events import *
from hikari.events.interaction_events import *
from hikari.events.lifetime_events import *
from hikari.events.member_events import *
from hikari.events.message_events import *
from hikari.events.monetization_events import *
from hikari.events.poll_events import *
from hikari.events.reaction_events import *
from hikari.events.role_events import *
from hikari.events.scheduled_events import *
from hikari.events.shard_events import *
from hikari.events.typing_events import *
from hikari.events.user_events import *
from hikari.events.voice_events import *
from hikari.files import URL as URL
from hikari.files import Bytes as Bytes
from hikari.files import File as File
from hikari.files import LazyByteIteratorish as LazyByteIteratorish
from hikari.files import Pathish as Pathish
from hikari.files import Rawish as Rawish
from hikari.files import Resourceish as Resourceish
from hikari.guilds import *
from hikari.impl import ClientCredentialsStrategy as ClientCredentialsStrategy
from hikari.impl import GatewayBot as GatewayBot
from hikari.impl import RESTApp as RESTApp
from hikari.impl import RESTBot as RESTBot
from hikari.intents import *
from hikari.interactions.base_interactions import *
from hikari.interactions.command_interactions import *
from hikari.interactions.component_interactions import *
from hikari.interactions.modal_interactions import *
from hikari.invites import *
from hikari.iterators import *
from hikari.locales import *
from hikari.messages import *
from hikari.monetization import *
from hikari.permissions import *
from hikari.polls import *
from hikari.presences import *
from hikari.scheduled_events import *
from hikari.sessions import *
from hikari.snowflakes import SearchableSnowflakeish as SearchableSnowflakeish
from hikari.snowflakes import SearchableSnowflakeishOr as SearchableSnowflakeishOr
from hikari.snowflakes import Snowflake as Snowflake
from hikari.snowflakes import Snowflakeish as Snowflakeish
from hikari.snowflakes import SnowflakeishOr as SnowflakeishOr
from hikari.snowflakes import SnowflakeishSequence as SnowflakeishSequence
from hikari.snowflakes import Unique as Unique
from hikari.stage_instances import *
from hikari.stickers import *
from hikari.templates import *
from hikari.traits import *
from hikari.undefined import UNDEFINED as UNDEFINED
from hikari.undefined import UndefinedNoneOr as UndefinedNoneOr
from hikari.undefined import UndefinedOr as UndefinedOr
from hikari.undefined import UndefinedType as UndefinedType
from hikari.users import *
from hikari.voices import *
from hikari.webhooks import *
